@@ -8,11 +8,13 @@ from csrbox.constants import *
 import csrbox.csrgrp_check as csrgrp_check
 import datetime
 import textwrap
+import math
 
 version = csrbox.__version__
 logger = logging.getLogger(__name__)
 
-def csr_gen (isa_file, grp_file, custom_file, custom_attr_file, work_dir, logging=False):
+def csr_gen (isa_file, grp_file, custom_file, custom_attr_file, work_dir,
+        probeinst, logging=False):
     '''
         Top level function which validates the sanity of the grouping yaml 
         followed by calls to subfunctions which generated the corresponding 
@@ -45,7 +47,7 @@ def csr_gen (isa_file, grp_file, custom_file, custom_attr_file, work_dir, loggin
     csr_define(isa_yaml , work_dir , 'csrbox.defines', logging)
     csr_decoder(isa_yaml, work_dir , 'csrbox_decoder.bsv', logging)
     csr_types(isa_yaml  , work_dir , 'csr_types.bsv', logging)
-    groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, logging)
+    groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging)
 
 def csr_types(isa_yaml, work_dir,  bsv, logging=False):
     '''
@@ -74,13 +76,17 @@ package {0} ;
     `include "csrbox.defines"
 
 '''.format(pkg_name))
-    bsvfile.write(csrreq_struct_temp.format(xlen))
+    if 'C' in isa_yaml['hart0']['ISA'] :
+        c_str = 'Bit#(1) pc_1;'
+    else:
+        c_str = ''
+    bsvfile.write(csrreq_struct_temp.format(xlen, c_str))
     bsvfile.write(func_temp.format(xlen))
     bsvfile.write(end_package_temp)
     bsvfile.close()
 
 
-def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, logging=False):
+def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, probeinst, logging=False):
     '''
         Function which co-ordinates the generation of all the bluespec code for
         the individual groups/stations and the top level bluespec module.
@@ -164,6 +170,35 @@ def groups_gen(isa_yaml, grp_yaml, attr_yaml, work_dir, logging=False):
     filename = 'csrbox.bsv'
     bsvfile = open(work_dir + filename, 'w')
     bsvfile.write(bsvstr)
+    bsvfile.close()
+
+    filename = 'csr_probe.bsv'
+    bsvfile = open(work_dir + filename, 'w')
+    bsvfile.write(csr_probe_function(csr_isa, attr_yaml, probeinst, logging))
+    bsvfile.close()
+
+
+def csr_probe_function(isa_yaml, attr_yaml,probeinst, logging=False):
+    bsvstr = ''
+    bsvstr += '''
+  function Bit#(XLEN) fn_probe_csr (Bit#(12) csr_addr);
+    case (csr_addr)
+'''
+    if logging:
+        logger.info('-- Creating Probe function ')
+    for csrname in isa_yaml:
+        if attr_yaml[csrname]['value_method']['required']:
+            bsvstr += '''
+        `{0} : return zeroExtend({2}.sbread.mv_csr_{1});'''.format(csrname.upper(),
+                csrname.lower(), probeinst)
+    bsvstr += '''
+    default: return 0;
+    endcase
+  endfunction
+'''
+    return bsvstr
+    
+
 
 def csr_decoder(isa_yaml, work_dir, bsv, logging=False):
 
@@ -189,14 +224,33 @@ def csr_decoder(isa_yaml, work_dir, bsv, logging=False):
     definefile.write(func_decoder)
 
     for node in isa_yaml['hart0']:
+        cond_str = ''
         if isinstance(isa_yaml['hart0'][node], dict):
             if 'address' in isa_yaml['hart0'][node] and ( isa_yaml['hart0'][node][rvxlen]['accessible'] == True) :
+                if 'S' in isa_yaml['hart0'][node]['priv_mode']:
+                    cond_str = 'if (misa[18] == 1)'
+                elif 'U' in isa_yaml['hart0'][node]['priv_mode'] and \
+                    isa_yaml['hart0'][node]['address'] in user_trap_csrs:
+                        cond_str = 'if ((misa[13]&misa[20]) == 1)'
+
                 csr_name = node
-                definefile.write(       case_address_temp.format(csr_name.upper()))
+                definefile.write(
+                        case_address_temp.format(csr_name.upper(), cond_str))
     definefile.write('\n           endcase')
     definefile.write('\n     return valid;')
     definefile.write('\n     endfunction')
 
+    definefile.write(func_csr_to_csr)
+    for node in isa_yaml['hart0']:
+        cond_str = ''
+        if isinstance(isa_yaml['hart0'][node], dict):
+            if 'address' in isa_yaml['hart0'][node] and ( isa_yaml['hart0'][node][rvxlen]['accessible'] == True) :
+                addr = int(isa_yaml['hart0'][node]['address'])
+                definefile.write(csr_str_temp.format(node.upper(), addr,
+                    node.lower()))
+    definefile.write('\n           endcase')
+    definefile.write('\n     endfunction')
+
     definefile.write(end_package_temp)
     definefile.close()
 
@@ -294,6 +348,25 @@ def get_reset_val(value, width, lsb, xlen):
     # convert extracted sub-string into decimal again
     return int(kBitSubStr, 2)
 
+def derive_wlrl_func(wlrl_list):
+
+    legal_list = []
+    for l in wlrl_list:
+        if ':' in l:
+            minval = int(l.split(':')[0],16)
+            maxval = int(l.split(':')[1],16)
+            legal_list += list(range(minval,maxval+1))
+
+    act_width = int(math.log2(max(legal_list)))
+
+    act_list = list(range(0,(2**act_width)+1))
+    if sorted(legal_list) == act_list:
+        full_wlrl = True
+    else:
+        full_wlrl = False
+
+    return full_wlrl, act_width
+    
 
 def derive_warl_func(field_name, width, warl, csr_dep):
     '''
@@ -913,6 +986,15 @@ def station_regs(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkg_name,
                             subcontent['type']['warl']['dependency_fields'])
                         bsvstr += temp_warlReg.format(subdescription,
                                     str(width), csrname+'_'+subfield, subreset_val, bsv_str)
+                    elif 'wlrl' in type_reg:
+                        full_wlrl, awidth = derive_wlrl_func(subcontent['type']['wlrl'])
+                        if full_wlrl:
+                            bsvstr +=  temp_register.format(subdescription,
+                                str(width), csrname+'_'+subfield, subreset_val)
+                        else:
+                            bsvstr += temp_wlrlReg.format(subdescription,
+                                    str(width), csrname+'_'+subfield, \
+                                            subreset_val, awidth)
                     else:
                         bsvstr += temp_readOnlyRegister.format(subdescription,
                                 str(width), csrname+'_'+subfield, 0)
@@ -1014,7 +1096,22 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     
     if logging:
         logger.debug('-- Creating interface')
+    # define a value method for each csr.
+    bsvstr += '''
+  interface Sbread;'''
+    for csrname in isa_yaml:
+        if logging:
+            logger.debug('-- Creating Interface for ' + csrname)
+        if attr_yaml[csrname]['value_method']['required']:
+            size = attr_yaml[csrname]['size']
+            bsvstr +=(methodv_dec_temp.format(size, csrname))
+    bsvstr += '''  
+  endinterface:Sbread'''
+
+
+
     bsvstr += interface_temp.format(pkgname)
+    bsvstr += '''    interface Sbread sbread;'''
     # define additional required interfaces
     req_regs = []
     for csrname in isa_yaml:
@@ -1030,14 +1127,6 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     if pmp_entries > 0:
         bsvstr += pmp_ifc.format(pmp_entries, physical_addr_sz)
 
-    # define a value method for each csr.
-    for csrname in isa_yaml:
-        if logging:
-            logger.debug('-- Creating Interface for ' + csrname)
-        if attr_yaml[csrname]['value_method']['required']:
-            size = attr_yaml[csrname]['size']
-            bsvstr +=(methodv_dec_temp.format(size, csrname))
-
     # define csr specific action methods.
     for csrname in isa_yaml:
         if logging:
@@ -1058,7 +1147,11 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     #--- create module
     if logging:
         logger.debug('-- Creating module')
-    bsvstr += module_temp.format(pkgname,'') + '\n'
+    mod_attr_str = ''
+    if 'minstret' in isa_yaml:
+        if isa_yaml['minstret']['rv'+str(xlen)]['accessible']:
+            mod_attr_str = '(*conflict_free="ma_incr_minstret,sbread.mv_csr_minstret"*)'
+    bsvstr += module_temp.format(pkgname,mod_attr_str) + '\n'
 
     if logging:
         logger.debug('-- Instatiating groups')
@@ -1227,10 +1320,14 @@ def csr_top(isa_yaml, grp_yaml, attr_yaml, rvxlen, pkgname, logging = False):
     # complete the value method for each csr.
     if logging:
         logger.debug('-- Defining value methods')
+    bsvstr +='''
+    interface Sbread sbread;'''
     for csrname in isa_yaml:
         if attr_yaml[csrname]['value_method']['required']:
             grp = find_group(grp_yaml, csrname.upper())
-            bsvstr += methodv_def_temp.format(csrname, grp+'.mv_csr_', csrname)
+            bsvstr += textwrap.indent(methodv_def_temp.format(csrname, grp+'.mv_csr_', csrname), '  ')
+    bsvstr +='''
+    endinterface'''
 
     
     # define csr specific action methods.
