@@ -20,7 +20,8 @@ import riscv_config.checker as riscv_config
 @click.option ('--customspec','-cspec', type=click.Path(exists=True, resolve_path=True, readable=True), help = "CUSTOM CSR File" )
 @click.option ('--customattr','-cattr', type=click.Path(exists=True, resolve_path=True, readable=True), help = "CUSTOM CSR File" )
 @click.option ('--workdir', default='./csrbox_work', type=click.Path(resolve_path=True, writable=True), help='Work directory path')
-def cli(verbose, isaspec, grpspec, customspec, customattr, workdir):
+@click.option ('--probeinst','-pinst',default='soc', type=str)
+def cli(verbose, isaspec, grpspec, customspec, customattr, workdir, probeinst):
     utils.setup_logging(verbose)
     logger = logging.getLogger()
     logger.handlers = []
@@ -58,7 +59,8 @@ def cli(verbose, isaspec, grpspec, customspec, customattr, workdir):
         logger.debug('Creating new bsv directory: ' + bsv_dir) 
         os.mkdir(bsv_dir)
 
-    csr_gen.csr_gen(isa_file, grpspec, custom_file, customattr, bsv_dir, logging=True)
+    csr_gen.csr_gen(isa_file, grpspec, custom_file, customattr, bsv_dir,
+            probeinst, logging=True)
 
 
  
